@@ -341,6 +341,7 @@ void	G_handle_reg_memb( configuration reg_memb, membership_id reg_memb_id )
                             if( grp->num_members == 0 )
 			    {
 				/* discard this empty group */
+                                sl_destruct ( &grp->MembersList, dispose);
                                 sl_remove (  &GroupsList, grp->name, dispose);
 				Num_groups--;
 				GlobalStatus.num_groups = Num_groups;
@@ -417,6 +418,7 @@ void	G_handle_reg_memb( configuration reg_memb, membership_id reg_memb_id )
 			    if( grp->num_members == 0 )
 			    {
 				/* discard this empty group */
+                                sl_destruct ( &grp->MembersList, dispose);
                                 sl_remove (  &GroupsList, grp->name, dispose);
 				Num_groups--;
 				GlobalStatus.num_groups = Num_groups;
@@ -658,6 +660,7 @@ void	G_handle_trans_memb( configuration trans_memb, membership_id trans_memb_id
 			if( grp->num_members == 0 )
 			{
 				/* discard this empty group */
+                                sl_destruct ( &grp->MembersList, dispose);
 				sl_remove ( &GroupsList, grp->name, dispose);
 				Num_groups--;
 				GlobalStatus.num_groups = Num_groups;
@@ -1117,6 +1120,7 @@ void	G_handle_leave( char *private_group_name, char *group_name )
 		if( grp->num_members == 0 )
 		{
 			/* discard this empty group */
+                        sl_destruct ( &grp->MembersList, dispose);
 		        sl_remove( &GroupsList, grp->name, dispose );
 			Num_groups--;
 			GlobalStatus.num_groups = Num_groups;
@@ -1286,6 +1290,7 @@ void	G_handle_kill( char *private_group_name )
 			grp->num_members--;
 			if( grp->num_members == 0 )
 			{
+                                sl_destruct ( &grp->MembersList, dispose);
                                 sl_remove( &GroupsList, grp->name, dispose );
 				Num_groups--;
 				GlobalStatus.num_groups = Num_groups;
@@ -1611,6 +1616,7 @@ static	void		G_compute_and_notify()
 		  orig_grp->num_members = orig_grp->MembersList.size;
 
 			/* free this Work group */
+                  sl_destruct(&currentgroup->MembersList, dispose);
 		  sl_remove(indices[i]->groups, currentgroup, dispose);
             }
 
@@ -1692,7 +1698,7 @@ static	void		G_compute_and_notify()
 		}       
 	}       
         
-	sl_remove_all( &work, dispose );        
+        sl_destruct( &work, dispose );
         
 	G_print();
 }
@@ -2245,7 +2251,7 @@ int	G_private_to_names( char *private_group_name, char *private_name, char *proc
 {
 	char	name[MAX_GROUP_NAME];
 	char	*pn, *prvn;
-	int	priv_name_len, proc_name_len;
+	unsigned int	priv_name_len, proc_name_len;
 	int	i,legal_private_name;
 
         memcpy(name, private_group_name, MAX_GROUP_NAME );
