@@ -88,8 +88,8 @@ public class SpreadConnection
 	// The Spread version.
 	//////////////////////
         private static final int SP_MAJOR_VERSION = 3;
-	private static final int SP_MINOR_VERSION = 16;
-	private static final int SP_PATCH_VERSION = 0;
+	private static final int SP_MINOR_VERSION = 17;
+	private static final int SP_PATCH_VERSION = 3;
 
         // The default authentication method
         ////////////////////////////////////
@@ -290,6 +290,34 @@ public class SpreadConnection
 		
 		return ((i0 << 24) | (i1 << 16) | (i2 << 8) | (i3));
 	}	
+
+    // Reads from inputsocket until all bytes read or exception raised
+    //////////////////////////////////////////////////////////////////
+    private void readBytesFromSocket(byte buffer[], String bufferTypeString) throws SpreadException
+    {
+	int byteIndex;
+	int rcode;
+	try
+	{
+	    for(byteIndex = 0 ; byteIndex < buffer.length ; byteIndex += rcode)
+	    {
+		rcode = socketInput.read(buffer, byteIndex, buffer.length - byteIndex);
+		if(rcode == -1)
+		{
+		    throw new SpreadException("Connection closed while reading " + bufferTypeString);
+		}
+	    }
+	}
+	catch(InterruptedIOException e) {
+	    throw new SpreadException("readBytesFromSocket(): InterruptedIOException " + e);
+	}
+	catch(IOException e)
+	{
+	    throw new SpreadException("readBytesFromSocket(): read() " + e);
+	}
+
+    }
+
 	
 	// Gets a string from an array of bytes.
 	////////////////////////////////////////
@@ -454,24 +482,13 @@ NOT SUPPORTED IN 1.1	*/
 		// Read the name.
 		/////////////////
 		byte buffer[] = new byte[len];
-		int numRead;
-		try
-		{
-			numRead = socketInput.read(buffer);
-		}
-		catch(IOException e)
-		{
-			throw new SpreadException("read(): " + e);
-		}
-		
+		readBytesFromSocket(buffer, "authname");
 		// System.out.println("readAuthMethods: string is " + new String(buffer));
 
-		// Check for not enough data.
-		/////////////////////////////
-		if(numRead != len)
-		{
-			throw new SpreadException("Connection clsoed during connect attempt to read authnames");
-		}
+		//		if(numRead != len)
+		//{
+		//	throw new SpreadException("Connection closed during connect attempt to read authnames");
+		//}
 		
 		// for now we ignore the list.
 		//////////////////////////////
@@ -658,23 +675,8 @@ NOT SUPPORTED IN 1.1	*/
 		// Read the name.
 		/////////////////
 		byte buffer[] = new byte[len];
-		int numRead;
-		try
-		{
-			numRead = socketInput.read(buffer);
-		}
-		catch(IOException e)
-		{
-			throw new SpreadException("read(): " + e);
-		}
-		
-		// Check for not enough data.
-		/////////////////////////////
-		if(numRead != len)
-		{
-			throw new SpreadException("Connection clsoed during connect attempt");
-		}
-		
+		readBytesFromSocket(buffer, "group name");
+
 		// Store the group.
 		///////////////////
 		group = new SpreadGroup(this, new String(buffer));
