@@ -1035,17 +1035,52 @@ void	Discard_packets()
 
 	if( Aru != Last_seq ) return;
 
+        /* Deliver packets that must be delivered before the transitional signal.
+         * Those up to the Aru for my old ring were delivered in Read_form2().
+         * So, it remains to deliver all packets up to the first hole or the first
+         * SAFE message. */
+        Alarmp( SPLOG_INFO, PROTOCOL,
+                "Discard_packets: delivering messages after old ring Aru before transitional\n" );
+
+        for( i = Last_discarded+1; i <= Highest_seq; i++ )
+        {
+            pack_entry = i & PACKET_MASK;
+	    if( ! Packets[pack_entry].exist )
+		Alarmp( SPLOG_FATAL, PROTOCOL, "Discard_packets: (EVS before transitional) packet %d not exist\n", i);
+	    if( Packets[pack_entry].exist == 3 )
+	    {
+		Alarmp( SPLOG_INFO, PROTOCOL, "Discard_packets: Found first Hole in %d\n", i);
+                break;
+	    }
+            if( Is_safe( Packets[pack_entry].head->type ) ) {
+                Alarmp( SPLOG_INFO, PROTOCOL, "Discard_packets: Found first SAFE message in %d", i);
+                break;
+            }
+            /* should deliver packet or dispose the body if it was delivered already */
+            if( Packets[pack_entry].exist == 1 ){
+                Deliver_packet( pack_entry, 0 );
+            } else {
+                dispose( Packets[pack_entry].body );
+            }
+            /* dispose packet header in any case */
+            dispose( Packets[pack_entry].head );
+            Alarmp( SPLOG_INFO, PROTOCOL, "Discard_packets: delivering %d in EVS\n",i);
+	    Packets[pack_entry].exist = 0;
+            Last_discarded = i;
+        }
+
 	/* calculate and deliver transitional membership */
+        Alarmp( SPLOG_INFO, PROTOCOL, "Discard_packets: Delivering transitional membership\n" );
 	Memb_transitional();
 	Sess_deliver_trans_memb( Trans_membership, Memb_trans_id() );
 
-	/* deliver all packets */
+	/* deliver all remaining packets for EVS */
 	found_hole = 0;
 	for( i = Last_discarded+1; i <= Highest_seq; i++ )
 	{
 	    pack_entry = i & PACKET_MASK;
 	    if( ! Packets[pack_entry].exist )
-		Alarm( EXIT, "Discard_packets: packet %d not exist\n",i);
+		Alarm( EXIT, "Discard_packets: (EVS after transitional) packet %d not exist\n", i);
 	    if( Packets[pack_entry].exist == 3 )
 	    {
 		/* 
@@ -1132,7 +1167,7 @@ void	Discard_packets()
 	{
 	    pack_entry = i & PACKET_MASK;
 	    if( ! Packets[pack_entry].exist )
-		Alarm( EXIT, "Discard_packets: packet %d not exist\n",i);
+		Alarm( EXIT, "Discard_packets: (NOT EVS) packet %d not exist\n",i);
 
 	    /* should deliver packet or dispose the body if it was delivered already */
 	    if( Packets[pack_entry].exist == 1 ) Deliver_packet( pack_entry, 0 );
@@ -1141,7 +1176,7 @@ void	Discard_packets()
 	    dispose( Packets[pack_entry].head );
 	    Packets[pack_entry].exist = 0;
 	}
-	Alarm( PROTOCOL, "Discard_packets: packets %d-%d where discarded\n",
+	Alarm( PROTOCOL, "Discard_packets: packets %d-%d were discarded\n",
 			Last_discarded+1, Aru );
 
 	Last_discarded = Aru;
