@@ -1014,7 +1014,7 @@ void    Sess_session_authorized(int ses)
 {
         char        ip[16];
         char        response;
-        int         name_len;
+        unsigned int    name_len;
         char	private_group_name[MAX_GROUP_NAME];
 
         if (!Is_preauth_session(Sessions[ses].status) )
